import ipywidgets as widgets
import pandas as pd
import plotly.express as px
from ipywidgets import interact


def plot(df: pd.DataFrame, points="all", barmode="group", **kwargs):
    """plot different chart types to compare performance and see performance variability across random seeds

    uses Plotly express; all available keyword arguments can be passed through to px.bar(), px.scatter(), etc.
    """

    options = list(df.columns) + [None]

    if "Incumbent" in options:
        # read custom axis data or use common defaults for summary DataFrame
        x_default = kwargs.pop("x", "Time")
        y_default = kwargs.pop("y", "Incumbent")
        color_default = kwargs.pop("color", "Log")
    else:
        # read custom axis data or use common defaults for summary DataFrame
        x_default = kwargs.pop("x", "Log")
        y_default = kwargs.pop("y", "Runtime")
        color_default = kwargs.pop("color", "Seed")

    # set type to None to disable plotting until a type is selected
    type_default = kwargs.pop("type", None)

    # pop values for remaining widgets to avoid double inputs
    symbol_default = kwargs.pop("symbol", None)
    log_x_default = kwargs.pop("log_x", False)
    log_y_default = kwargs.pop("log_y", False)

    # check wether selected keys are available in DataFrame
    if x_default not in options:
        x_default = None
    if y_default not in options:
        y_default = None
    if color_default not in options:
        color_default = None

    switches = dict(
        x=widgets.Dropdown(options=options, value=x_default),
        y=widgets.Dropdown(options=options, value=y_default),
        color=widgets.Dropdown(options=options, value=color_default),
        type=widgets.Dropdown(
            options=["box", "bar", "scatter", "line"], value=type_default
        ),
        symbol=widgets.Dropdown(options=options, value=symbol_default),
        log_x=widgets.Checkbox(value=log_x_default),
        log_y=widgets.Checkbox(value=log_y_default),
    )

    @interact(**switches)
    def _(x, y, color, type, symbol, log_x, log_y):
        if type == "box":
            return px.box(
                df,
                x=x,
                y=y,
                color=color,
                points=points,
                log_x=log_x,
                log_y=log_y,
                **kwargs,
            )
        elif type == "bar":
            return px.bar(
                df,
                x=x,
                y=y,
                color=color,
                barmode=barmode,
                log_x=log_x,
                log_y=log_y,
                **kwargs,
            )
        elif type == "scatter":
            return px.scatter(
                df,
                x=x,
                y=y,
                color=color,
                symbol=symbol,
                log_x=log_x,
                log_y=log_y,
                **kwargs,
            )
        elif type == "line":
            return px.line(
                df,
                x=x,
                y=y,
                color=color,
                symbol=symbol,
                log_x=log_x,
                log_y=log_y,
                **kwargs,
            )



import re

from grblogtools.parsers.barrier import BarrierParser
from grblogtools.parsers.simplex import SimplexParser
from grblogtools.parsers.util import typeconvert_groupdict


class ContinuousParser:
    # The pattern indicating the start or the termination of the barrier/simplex
    # in case of solving a MIP. In some cases, the log might only include this one
    # line
    mip_relaxation_pattern = re.compile(
        r"Root relaxation: objective (?P<RelaxObj>[^,]+), (?P<RelaxIterCount>\d+) iterations, (?P<RelaxTime>[^\s]+) seconds"
    )

    barrier_interruption_pattern = re.compile(
        r"Barrier solve interrupted - model solved by another algorithm"
    )

    continuous_termination_patterns = [
        re.compile(
            r"(?P<SUBOPTIMAL>Sub-optimal termination)(?: - objective (?P<ObjVal>.*))$"
        ),
        re.compile(r"(?P<OPTIMAL>Optimal objective\s+(?P<ObjVal>.*))$"),
    ]

    def __init__(self):
        """Initialize the Continuous parser."""
        self._barrier_parser = BarrierParser()
        self._simplex_parser = SimplexParser()

        self._summary = {}

        self._current_pattern = None

    def parse(self, line: str) -> bool:
        """Parse the given log line to populate summary and progress data.

        It defers to the simplex and the barrier parsers as needed.

        Args:
            line (str): A line in the log file.

        Returns:
            bool: Return True if the given line is matched by some pattern.
        """
        mip_relaxation_match = ContinuousParser.mip_relaxation_pattern.match(line)
        if mip_relaxation_match:
            self._current_pattern = "relaxation"
            self._summary.update(typeconvert_groupdict(mip_relaxation_match))
            return True

        for pattern in ContinuousParser.continuous_termination_patterns:
            match = pattern.match(line)
            if match:
                for key, value in typeconvert_groupdict(match).items():
                    if key in ["OPTIMAL", "SUBOPTIMAL"]:
                        self._summary.update({"Status": key})
                    else:
                        self._summary.update({key: value})
                return True

        if self._current_pattern is None:
            if self._barrier_parser.parse(line):
                self._current_pattern = "barrier"
                return True
            if self._simplex_parser.parse(line):
                self._current_pattern = "simplex"
                return True

            return False

        if self._current_pattern == "barrier":
            matched = self._barrier_parser.parse(line)
            # If the barrier gets interrupted during the concurrent or there are
            # extra simplex iterations, switch to simplex
            if not matched and (
                ContinuousParser.barrier_interruption_pattern.match(line)
                or self._simplex_parser.parse(line)
            ):
                self._current_pattern = "simplex"
                return True
            return matched

        if self._current_pattern == "simplex":
            match = self._simplex_parser.parse(line)
            return match

        return False

    def get_summary(self) -> dict:
        """Return the current parsed summary."""
        self._summary.update(self._barrier_parser.get_summary())
        self._summary.update(self._simplex_parser.get_summary())
        return self._summary

    def get_progress(self) -> list:
        """Return the detailed progress in the continuous method."""
        return self._barrier_parser.get_progress() + self._simplex_parser.get_progress()



import grblogtools as glt


def test_get_dataframe():
    """Just check we read without errors."""
    summary = glt.get_dataframe(["data/*.log"])


def test_read_with_timelines():
    """Check timelines argument."""
    summary, timelines = glt.get_dataframe(["data/*.log"], timelines=True)
    assert set(timelines.keys()) == {"norel", "rootlp", "nodelog"}


def test_norel_timeline():
    """Check norel logs before the root node. Note that this reports norel's
    clock, so read + presolve time would need to be added for 'real' time."""
    _, timelines = glt.get_dataframe(
        ["tests/assets/mip_norel.log"],
        timelines=True,
    )
    norel = timelines["norel"]
    assert set(norel.columns) == {
        "Time",
        "BestBd",
        "Incumbent",
        "Log",
        "LogFilePath",
        "Model",
        "ModelFile",
        "Version",
        "LogNumber",
        "ModelFilePath",
        "Seed",
    }
    assert norel.shape[0] == 15
    assert len(norel["LogFilePath"].unique()) == 1
    assert norel["Time"].min() == 5.0
    assert norel["Time"].max() == 93.0
    assert (norel["Incumbent"].max() - 1.450014e09) <= 1e05
    assert (norel["Incumbent"].min() - 1.2000e09) <= 1e05
    assert (norel["BestBd"].max() - 8.00002e08) <= 1e05
    assert (norel["BestBd"].min() - 8.00002e08) <= 1e05
    # Sometimes a bound isn't reported.
    assert norel["BestBd"].isnull().sum() == 1


def test_newsolution_markers():
    _, timelines = glt.get_dataframe(["data/912-glass4-1.log"], timelines=True)
    new_solutions = timelines["nodelog"]["NewSolution"]
    assert new_solutions.value_counts().to_dict() == {"H": 19, "*": 4}
